@@ -85,7 +85,6 @@
 #include <linux/mroute.h>
 #include <linux/netfilter_ipv4.h>
 #include <linux/random.h>
-#include <linux/jhash.h>
 #include <net/protocol.h>
 #include <net/ip.h>
 #include <net/route.h>
@@ -114,20 +113,17 @@ int ip_rt_redirect_load		= HZ / 50;
 int ip_rt_redirect_silence	= ((HZ / 50) << (9 + 1));
 int ip_rt_error_cost		= HZ;
 int ip_rt_error_burst		= 5 * HZ;
-/* SpeedMod: Tune ip_rt_gc_elasticity */
-//int ip_rt_gc_elasticity		= 8;
-int ip_rt_gc_elasticity		= 1;
+int ip_rt_gc_elasticity		= 8;
 int ip_rt_mtu_expires		= 10 * 60 * HZ;
 int ip_rt_min_pmtu		= 512 + 20 + 20;
 int ip_rt_min_advmss		= 256;
-int ip_rt_secret_interval	= 10 * 60 * HZ;
+
 static unsigned long rt_deadline;
 
 #define RTprint(a...)	printk(KERN_DEBUG a)
 
 static struct timer_list rt_flush_timer;
 static struct timer_list rt_periodic_timer;
-static struct timer_list rt_secret_timer;
 
 /*
  *	Interface to generic destination cache.
@@ -198,17 +194,19 @@ struct rt_hash_bucket {
 static struct rt_hash_bucket 	*rt_hash_table;
 static unsigned			rt_hash_mask;
 static int			rt_hash_log;
-static unsigned int		rt_hash_rnd;
 
 struct rt_cache_stat rt_cache_stat[NR_CPUS];
 
 static int rt_intern_hash(unsigned hash, struct rtable *rth,
 				struct rtable **res);
 
-static unsigned int rt_hash_code(u32 daddr, u32 saddr, u8 tos)
+static __inline__ unsigned rt_hash_code(u32 daddr, u32 saddr, u8 tos)
 {
-	return (jhash_3words(daddr, saddr, (u32) tos, rt_hash_rnd)
-		& rt_hash_mask);
+	unsigned hash = ((daddr & 0xF0F0F0F0) >> 4) |
+			((daddr & 0x0F0F0F0F) << 4);
+	hash ^= saddr ^ tos;
+	hash ^= (hash >> 16);
+	return (hash ^ (hash >> 8)) & rt_hash_mask;
 }
 
 static int rt_cache_get_info(char *buffer, char **start, off_t offset,
@@ -423,8 +421,6 @@ static void SMP_TIMER_NAME(rt_run_flush)(unsigned long dummy)
 
 	rt_deadline = 0;
 
-	get_random_bytes(&rt_hash_rnd, 4);
-
 	for (i = rt_hash_mask; i >= 0; i--) {
 		write_lock_bh(&rt_hash_table[i].lock);
 		rth = rt_hash_table[i].chain;
@@ -483,14 +479,6 @@ void rt_cache_flush(int delay)
 	spin_unlock_bh(&rt_flush_lock);
 }
 
-static void rt_secret_rebuild(unsigned long dummy)
-{
-	unsigned long now = jiffies;
-
-	rt_cache_flush(0);
-	mod_timer(&rt_secret_timer, now + ip_rt_secret_interval);
-}
-
 /*
    Short description of GC goals.
 
@@ -2409,15 +2397,6 @@ ctl_table ipv4_route_table[] = {
 		maxlen:		sizeof(int),
 		mode:		0644,
 		proc_handler:	&proc_dointvec,
-	},
-	{
-		ctl_name:	NET_IPV4_ROUTE_SECRET_INTERVAL,
-		procname:	"secret_interval",
-		data:		&ip_rt_secret_interval,
-		maxlen:		sizeof(int),
-		mode:		0644,
-		proc_handler:	&proc_dointvec_jiffies,
-		strategy:	&sysctl_jiffies,
 	},
 	 { 0 }
 };
@@ -2449,25 +2428,15 @@ static int ip_rt_acct_read(char *buffer, char **start, off_t offset,
 		*eof = 1;
 	}
 
-	offset /= sizeof(u32);
+	/* Copy first cpu. */
+	*start = buffer;
+	memcpy(buffer, IP_RT_ACCT_CPU(0), length);
 
-	if (length > 0) {
-		u32 *src = ((u32 *) IP_RT_ACCT_CPU(0)) + offset;
-		u32 *dst = (u32 *) buffer;
-
-		/* Copy first cpu. */
-		*start = buffer;
-		memcpy(dst, src, length);
-
-		/* Add the other cpus in, one int at a time */
-		for (i = 1; i < smp_num_cpus; i++) {
-			unsigned int j;
-
-			src = ((u32 *) IP_RT_ACCT_CPU(i)) + offset;
-
-			for (j = 0; j < length/4; j++)
-				dst[j] += src[j];
-		}
+	/* Add the other cpus in, one int at a time */
+	for (i = 1; i < smp_num_cpus; i++) {
+		unsigned int j;
+		for (j = 0; j < length/4; j++)
+			((u32*)buffer)[j] += ((u32*)IP_RT_ACCT_CPU(i))[j];
 	}
 	return length;
 }
@@ -2477,9 +2446,6 @@ void __init ip_rt_init(void)
 {
 	int i, order, goal;
 
-	rt_hash_rnd = (int) ((num_physpages ^ (num_physpages>>8)) ^
-			     (jiffies ^ (jiffies >> 7)));
-
 #ifdef CONFIG_NET_CLS_ROUTE
 	for (order = 0;
 	     (PAGE_SIZE << order) < 256 * sizeof(struct ip_rt_acct) * NR_CPUS; order++)
@@ -2498,9 +2464,7 @@ void __init ip_rt_init(void)
 	if (!ipv4_dst_ops.kmem_cachep)
 		panic("IP: failed to allocate ip_dst_cache\n");
 
-	/* SpeedMod: goal=32 gives 16384 buckets for 4K page size */
-	goal = 32;
-//	goal = num_physpages >> (26 - PAGE_SHIFT);
+	goal = num_physpages >> (26 - PAGE_SHIFT);
 //	goal = num_physpages >> (21 - PAGE_SHIFT);
 
 	for (order = 0; (1UL << order) < goal; order++)
@@ -2534,13 +2498,8 @@ void __init ip_rt_init(void)
 //	ip_rt_max_size = (rt_hash_mask + 1) * 2;
 //	ipv4_dst_ops.gc_thresh = (ip_rt_max_size / 4);
 
-	/* SpeedMod: Tuning */
-/*
 	ipv4_dst_ops.gc_thresh = (rt_hash_mask + 1);
 	ip_rt_max_size = (rt_hash_mask + 1) * 16;
-*/
-	ipv4_dst_ops.gc_thresh = (rt_hash_mask + 1);
-	ip_rt_max_size = (rt_hash_mask + 1) * 2;
 
 //	printk("gc_thresh=%d\n", ipv4_dst_ops.gc_thresh);
 //	printk("ip_rt_max_size=%d\n", ip_rt_max_size);
@@ -2553,7 +2512,6 @@ void __init ip_rt_init(void)
 
 	rt_flush_timer.function = rt_run_flush;
 	rt_periodic_timer.function = rt_check_expire;
-	rt_secret_timer.function = rt_secret_rebuild;
 
 	/* All the timers, started at system startup tend
 	   to synchronize. Perturb it a bit.
@@ -2562,10 +2520,6 @@ void __init ip_rt_init(void)
 					ip_rt_gc_interval;
 	add_timer(&rt_periodic_timer);
 
-	rt_secret_timer.expires = jiffies + net_random() % ip_rt_secret_interval +
-		ip_rt_secret_interval;
-	add_timer(&rt_secret_timer);
-
 	proc_net_create ("rt_cache", 0, rt_cache_get_info);
 	proc_net_create ("rt_cache_stat", 0, rt_cache_stat_get_info);
 #ifdef CONFIG_NET_CLS_ROUTE
